{% if messages[0]['role'] == 'system' %}{% set loop_messages = messages[1:] %}{% set system_message = messages[0]['content'].strip() + '\n\n' %}{% else %}{% set loop_messages = messages %}{% set system_message = '' %}{% endif %}{{ system_message }}{% for message in loop_messages %}{% if (message['role'] == 'user') != (loop.index0 % 2 == 0) %}{{ raise_exception('Conversation roles must alternate user/assistant/user/assistant/...') }}{% endif %}{% if message['role'] == 'user' %}{{ 'USER: ' + message['content'].strip()}}{% elif message['role'] == 'assistant' %}{{ ' ASSISTANT: ' + message['content'].strip() + '</s>'}}{% endif %}{% if loop.last and message['role'] == 'user' and add_generation_prompt %}{{ ' ASSISTANT: ' }}{% endif %}{% endfor %}